#!/usr/local/bin/perl
#
# PPML ver 0.75 --- "Pigeon Post" Mail-list Driver
#                ( 1999.07.02 named by Masato MINDA <minmin@wide.ad.jp> )
# Copyright (c)1999 by Kazunori ANDO <ando@kk.iij4u.or.jp> all rights reserved.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
################
$PPML_VERSION = "0.751";
################
use IO::Socket;
$DIR = $ARGV[0];
if ( $0 =~ /^(\S+)\/ppml\.pl/ ){ $ORIGIN = $1 };
$LOCK_SH = 1;
$LOCK_EX = 2;
$LOCK_NB = 4;
$LOCK_UN = 8;
$locktmp = "L$$";
#################
## exit status ##
#################
$EX_OK         =  0; # successful termination
$EX__BASE      = 64; # base value for error messages
$EX_USAGE      = 64; # command line usage error
$EX_DATAERR    = 65; # data format error
$EX_NOINPUT    = 66; # cannot open input
$EX_NOUSER     = 67; # addressee unknown
$EX_NOHOST     = 68; # host name unknown
$EX_UNAVAILABLE= 69; # service unavailable
$EX_SOFTWARE   = 70; # internal software error
$EX_OSERR      = 71; # system error (e.g., can't fork)
$EX_OSFILE     = 72; # critical OS file missing
$EX_CANTCREAT  = 73; # can't create (user) output file
$EX_IOERR      = 74; # input/output error
$EX_TEMPFAIL   = 75; # temp failure; user is invited to retry
$EX_PROTOCOL   = 76; # remote error in protocol
$EX_NOPERM     = 77; # permission denied
$EX_CONFIG     = 78; # configuration error
$EX__MAX       = 78; # maximum listed value
##### main #####
&ReadSiteConfig;
&ChdirByArg;         # Chdir to $DIR
&ReadConfig;         # Read config.ph
&ParseInput(*e);     # Read input and set %e hash
&AuthPoster(*e);     # Poster authorization by address
&PPLockWait;         # Lock
&CheckCache(*e);     # Loop check by body-checksum and Message-Id
&IncrSeqNum(*e);     # Increment sequence number
&Unlock;
&MIMEdecode(*e);     # MIME decode for headers
&MakeHeader(*e);     # Delete,change and add headers
#&DebugPrint(*e);
&MIMEencode(*e);     # MIME encode for headers
&SaveAsFile(*e) unless ($NOARCHIVE);     # Article saved to $Savedir
&Distribute(*e);     # Distribute mail by SMTP session
&Done;               # Done
################

sub ReadSiteConfig{
    $0 = "ReadSiteConfig";
    unshift(@INC,$ORIGIN);
    eval("require(\"sitedef.ph\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
	exit $EX_CONFIG;
    }
    return 1;
}

sub ChdirByArg{
    $0 = "ChdirByArg";
    if ((-d $DIR) && (-x $DIR)){
	unshift(@INC, $DIR);
	chdir $DIR;
    }else{
	print STDERR "$0:WARN: \$DIR is not given or Can't chdir.";
	exit $EX_NOPERM;
    }
    return 1;
}

sub ReadConfig{
    $0 = "ReadConfig";
    eval("require(\"config.ph\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
	&Unlock;
        exit $EX_CONFIG;
    }
    return 1;
}

sub ParseInput{
    $0 = "ParseInput";
    local(*e) = @_;
    local(@header);
    local($mpp,$mpi,$h,$c,$hc,$s);
    $/ = "\n\n";
    $e{'Header'} = <STDIN>;
    $/ = "\n";
    $e{'size:header'} = length($e{'Header'});
    @header = split(/\n/,$e{'Header'});
    foreach (@header){
	if (/^([^\s\:]+)\:\s*(.*)/){
	    $h = $1;
	    $c = $2;
	    # field-name covert to unified form
	    $horg = $h;
	    $h =~ tr/A-Z/a-z/;
	    # message-id -> Message-id
	    if ( $h =~ /^(.)(.*)/ ){
                $hc = $1;
                $hc =~ tr/a-z/A-Z/;
                $h = $hc.$2;
	    }
	    # Message-id -> Message-Id
	    while ( $h =~ /^(.+)(\-[a-z])(.+)/ ){
                $hc = $2;
                $hc =~ tr/a-z/A-Z/;
                $h = $1.$hc.$3;
	    }
	    $e{"OH:$h"} = $horg;
	    if ($h =~ /Received/){
		$recn++;
	    }
	    if ((length($e{"H:$h"}) > 0)&&(length($c) > 0)){
		$e{"H:$h"} .= "$_\n";
	    }elsif((length($h) > 0) && (length($c) > 0)){ 
		$e{"H:$h"} .= "$c\n";
	    }
	}elsif(/^(\s+.*)/){
	    if ($h ne ""){
		$e{"H:$h"} .= "$_\n";
	    }else{
		$e{'bogus'} .= "$_\n";
	    }
	}else{
	    $e{'bogus'} .= "$_\n";
	}
    }
    # Save Received header splited...
    $e{"num:received"} = $recn;
    $recn--;
    foreach $reci (split(/\n/,$e{"H:Received"})){
	$reci =~ s/\n//g;
	if ($reci =~ /^(Received\:)\s*(.*)$/){
	    $recn--;
	    $e{"received:$recn"} = "$2\n";
	}else{
	    $e{"received:$recn"} .= "$reci\n";
	}
    }
    while(<STDIN>){
	$e{'Body'} .= $_;
	$e{'line:body'}++;
	$e{'size:body'} += length($_);
    }
    if ($CHKSUMTYPE eq 'SHA1') {
	use Digest::SHA1 qw(sha1_base64);
	$e{'checksum:body'} = sha1_base64($e{'Body'});
    }elsif($CHKSUMTYPE eq 'MD5'){
	use Digest::MD5  qw(md5_base64);
	$e{'checksum:body'} = md5_base64($e{'Body'});
    }else{
	$e{'checksum:body'} = sprintf("%011d-%011d",$e{'size:body'},unpack("%31C*",$e{'Body'}));
    }
    $e{'size'} = $e{'size:header'} + $e{'size:body'};
    &Log("$0:FROM: $e{'H:From'}");
    &Log("$0:DATA: header:$e{'size:header'} + body:$e{'size:body'} = $e{'size'} bytes");
    # multipart analyze
    if ((defined($MAXPARTSIZE)||$MIMEANALYZE) &&
	(($e{'H:Content-Type'} =~ /[\"]*multipart\/mixed[\"]*[\;]*/i)|| 
	 ($e{'H:Content-Type'} =~ /[\"]*multipart\/related[\"]*[\;]*/i)||
         ($e{'H:Content-Type'} =~ /[\"]*multipart\/alternative[\"]*[\;]*/i) )&&
	($e{'H:Content-Type'} =~ /boundary\=[\"]*([^\"\;\s]+)[\"]*/i)){ #"
	$e{'mp:00:boundary'} = $1;
	$e{'size:mp:00:boundary'} = length($e{'mp:00:boundary'});
	# partscan
        if ($e{'mp:00:boundary'} ne ""){
	    &ParseParts(0,0,$e{'Body'},$e{'mp:00:boundary'},*e);
        }
    }
    &CheckVirus(*e);
    return 1;
}

sub ParseParts{
    $0 = "ParseParts";
    local($b,$mp,$part,$bound,*e) = @_;
    local($mcs) = sprintf("mp:%02d.%02d",$mp,0);
    local($mcb) = sprintf("mp:%02d",$mp);
    $e{"last:$mcb"} = 0;
    local($mc) = 0;
    local($i);
    chomp($bound);
    foreach $line (split(/\n/,$part)){
	if ($line eq "--$bound--"){
	    $e{"last:$mcb"} = $mc;
	    last;
        }
	if ($line eq "--$bound"){
	    $mc++;
	    $mcs = sprintf("mp:%02d.%02d",$mp,$mc);
	    next;
	}
	$e{$mcs} .= "$line\n";
    }
    if ($e{"last:$mcb"} == 0){
	$e{"last:$mcb"} = $mc;
    }
    $e{$mcb} = $mc;
    &Log("$0:MIME: multipart $mc parts in $mcb message(bound=$bound)");
    for ($i = 1; $i <= $mc ; $i++){
	$mcs = sprintf("mp:%02d.%02d",$mp,$i);
	### MIME analyze
	$e{"size:$mcs"} = length($e{$mcs});
	$e{"mhead:$mcs"} = substr($e{$mcs},0,index($e{$mcs},"\n\n")+1);
	$e{"mbody:$mcs"} = substr($e{$mcs},index($e{$mcs},"\n\n")+2,index($e{$mcs},"--$bound"));
	## mime body
	if ($CHKSUMTYPE eq 'SHA1') {
	    use Digest::SHA1 qw(sha1_base64);
	    $e{"checksum:$mcs"} = sha1_base64($e{"mbody:$mcs"});
	}elsif($CHKSUMTYPE eq 'MD5'){
	    use Digest::MD5  qw(md5_base64);
	    $e{"checksum:$mcs"} = md5_base64($e{"mbody:$mcs"});
	}else{
	    $e{"checksum:$mcs"} = sprintf("%011d-%011d",$e{"size:$mcs"},unpack("%31C*",$e{"mbody:$mcs"}));
	}
	$s = "\'".$e{"checksum:$mcs"}."\':";
	## mime header
	if ( $e{"mhead:$mcs"} =~ /Content\-Type\:\s*([^\;]+)\s*\;/i){
	    $e{"mtype:$mcs"} = lc($1);
	    if ((($e{"mtype:$mcs"} =~ /multipart\/mixed/i)||
		 ($e{"mtype:$mcs"} =~ /audio\/x-wav/i)||
		 ($e{"mtype:$mcs"} =~ /multipart\/alternative/i)) &&
		($e{"mhead:$mcs"} =~ /boundary\=[\"]*([^\"\;\s]+)[\"]*/i)){ #"
		$e{"$mcs:boundary"} = $1;
		$e{"size:$mcs:boundary"} = length($e{"$mcs:boundary"});
	    }
	}else{
	    $e{"mtype:$mcs"} = 'unknown';
	}
	$s .= $e{"mtype:$mcs"}.":";
	if ( $e{"mhead:$mcs"} =~ /charset\=([^\;]+)\;/i){ 
	    $e{"mchar:$mcs"} = lc($1);
	}else{
	    $e{"mchar:$mcs"} = 'unknown';
	}
	$s .= $e{"mchar:$mcs"}.":";
	if ( $e{"mhead:$mcs"} =~ /Content-Transfer-Encoding\:\s*(\S+)/i){
	    $e{"mcode:$mcs"} = lc($1);
	}else{
	    $e{"mcode:$mcs"} = 'unknown';
	}
	$s .= $e{"mcode:$mcs"}.":";
	if ( $e{"mhead:$mcs"} =~ /name\=\"*([^\"\n]+)\"*/i){ #"
	    $e{"mname:$mcs"} = $1;
	}elsif( $e{"mhead:$mcs"} =~ /filename\=\"*([^\"\n]+)\"*/i){ #"
	    $e{"mname:$mcs"} = $1;
	}else{
	    $e{"mname:$mcs"} = 'unknown';
	}
	$s .= $e{"mname:$mcs"};
	&Log("$0:MIME:$mcs:$s");
	if ( ($b == 0) && ($e{"$mcs:boundary"} ne "") ){
	    &ParseParts(1,$i,$e{$mcs},$e{"$mcs:boundary"},*e);
	}
    }
    return 1;
}


sub CheckVirus{
    $0 = "CheckVirus";
    local(*e) = @_;
    local(%vpart,%vbody);
    ## ppvirus.ph spec --
    ## %vpart = ('key','name'); %vbody = ('key','name');
    eval("require(\"ppvirus.ph\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
	exit $EX_CONFIG;
    }
    if ( defined($vbody{$e{'checksum:body'}})){
	&Log("$0:Fail: VIRUS check (Body: $vbody{$e{'checksum:body'}})");
	&Log("$0:Done: Article rejected (VIRUS)");
	if ($VIRUSNOTIFY){
	    &NoticeMail($ML_ADMIN,"VIRUS check by PPML($BRACKET)",$e{'Header'}.$e{'Body'},*r);
	}
	&Done;
    }
    for ($b = 0; $b <= $e{'mp:00'}; $b++){
	$mcb = sprintf("mp:%02d",$b);
	for ($i = 1; $i <= $e{$mcb}; $i++){
	    $mcs = sprintf("mp:%02d.%02d",$b,$i);
	    if ( defined($vpart{$e{"checksum:$mcs"}}) ){
		$vname = $vpart{$e{"checksum:$mcs"}};
		&Log("$0:Fail: VIRUS check ($mcs: $vname)");
		&Log("$0:Done: Article rejected (VIRUS)");
		if ($VIRUSNOTIFY){
		    &NoticeMail($ML_ADMIN,"VIRUS check by PPML($BRACKET)",$e{'Header'}.$e{'Body'},*r);
		}
		&Done;
	    }
	}
    }
    &Log("$0:PASS: $e{'checksum:body'}");
    return 1;
}


sub CheckCache{
    $0 = "CheckCache";
    local(*e) = @_;
    local(%mc,%cm);
    local($mid);
    $mid = $e{'H:Message-Id'} || $e{'H:Message-ID'};
    $mid =~ s/\r//;
    $mid =~ s/\n//;
    if (-f $Cache){
	open(CACHE,$Cache);
	while (<CACHE>){
	    if (/^(\S+)\s+(\<[^\>]+\>)\s*(\d*)$/){
		$cm{$1} = $2;
		$mc{$2} = $1;
	    }
	}
	close(CACHE);
	if ($DISCARD_LOOP){
	    if ($cm{$e{'checksum:body'}}) {
		&Log("$0:Fail: Loop check(checksum:$e{'checksum:body'})");
		&NoticeMail($ML_ADMIN,"Loop check($BRACKET)",$e{'Header'}.$e{'Body'},*r);
		&Log("$0:Done: Article discarded and forwarded to ML_ADMIN");

		&Unlock;
                &Done;
	    }
	    if ($mc{$mid}) {
		&Log("$0:Fail: Loop check(Message-Id:$mid)");
		&NoticeMail($ML_ADMIN,"Loop check($BRACKET)",$e{'Header'}.$e{'Body'},*r);
		&Log("$0:Done: Article discarded and forwarded to ML_ADMIN");
		&Unlock;
                &Done;
	    }
	}else{
	    if ($cm{$e{'checksum:body'}}) {
		&Log("$0:Fail: Loop check(checksum:$e{'checksum:body'})");
		print STDERR "Fail: Loop check\n";
		&Log("$0:Done: Article rejected");
		&Unlock;
		exit $EX_UNAVAILABLE;
	    }
	    if ($mc{$mid}) {
		&Log("$0:Fail: Loop check(Message-Id:$mid)");
		print STDERR "Fail: Loop check\n";
		&Log("$0:Done: Article rejected");
		&Unlock;
		exit $EX_UNAVAILABLE;
	    }
	}
    }
    open(CACHE,">>$Cache");
    print CACHE "$e{'checksum:body'} $mid\n";
    close(CACHE);
    &Log("$0:PASS: $e{'checksum:body'} $mid");

    return 1;
}

sub AuthPoster{
    $0 = "AuthPoster";
    local(*e) = @_;
    local(@ret);
    local($result);
    if (!defined(@MEMBERLIST)){
	&Log("$0:PASS: NOCHECK (not defined \@MEMBERLIST)");
	return 1;
    }
    &Conv2MBs($e{'H:From'},*ret);
    &Conv2MBs($e{'H:Sender'},*ret);
    &Conv2MBs($e{'H:Return-Path'},*ret);
    &Conv2MBs($e{'H:Resent-From'},*ret);
#    &Conv2MBs($e{'H:Reply-To'},*ret);
#   foreach $dom (@AUTHDOMAINS){
#	foreach $addr (@ret){
#	    if ($addr =~ /\@$dom$/i){
#		$e{'auth'} = 3;
#		$e{'poster'} = $ret[0];
#		return 1;
#	    }
#	}
#   }
    $e{'auth'} = &IsAuthPoster(*MEMBERLIST,*ret,*e);
    if ($e{'auth'} == 1){
	return 1;
    }elsif($e{'auth'} == 2){
	return 1;
    }else{
	&Log("$0:Fail: Not in members : @ret");
	&NoticeMail($ML_ADMIN,"NOT MEMBER($BRACKET)",$e{'Header'}.$e{'Body'},*r);
	&Log("$0:Done: Notice sent to \$ML_ADMIN($ML_ADMIN) -- notice done.");
        if ($NOTICE_TO_POSTER){
	    $return = $ret[0];
	    &NoticeMail($return,"NOT MEMBER($BRACKET)",$e{'Header'}.$e{'Body'},*r);
	    &Log("$0:Done: Notice sent to poster($return) -- notice done.");
	}
	&Done;
    }
}    

sub IsAuthPoster{
    $0 = "IsAuthPoster";
    local(*target,*ret,*e) = @_;
    local($ml,$domain) = split(/\@/,$ML_ADDR);
    foreach $list (@target){
	delete $INC{$list};
	undef %members;
	eval("require(\"$list\")");
	if ($@) {
	    &Log("$0:ERROR:$@");
	    print STDERR $@;
	    exit $EX_CONFIG;
	}
	if ($OPTIMIZED_DELIVERLIST == 0){
	    foreach $addr (@ret){
		$addr = &FormatAddr($addr);
		$lcaddr = &LcAddr($addr);
		if (defined($members{$addr})){
		    $e{'poster'} = $addr;
		    &Log("$0:AUTH: $addr($members{$addr}) found in $list.");
		    return 1;
		}elsif(defined($members{"\# $addr"})){
		    $e{'poster'} = $addr;
		    &Log("$0:AUTH: $addr found in $list(off).");
		    return 2;
		}elsif(defined($members{"$lcaddr"})){
		    $e{'poster'} = $addr;
		    &Log("$0:AUTH: $addr found in $list(old format).");
		    return 1;
		}elsif(defined($members{"\# $lcaddr"})){
		    $e{'poster'} = $addr;
		    &Log("$0:AUTH: $addr found in $list(off/old format).");
		    return 1;
		}
	    }
	}else{
	    foreach $addr (@ret){
		$addr = &FormatAddr($addr);
		$lcaddr = &LcAddr($addr);
		my($exist) = 0;
		foreach $mem (keys(%members)){
		    $exist = 1 if ($mem eq $addr);
		    $exist = 2 if ($mem eq "# $addr");
		    $exist = 3 if ($mem eq $lcaddr);
		    $exist = 4 if ($mem eq "# $lcaddr");
		}
		if ($exist == 1){
		    if ($members{$addr}[1] ne ""){
			for ($i = 1; $members{$addr}[$i] ne ""; $i++){
			    $m = $members{$addr}[$i];
			    if ($m eq $ml){
				$e{'poster'} = $addr;
				&Log("$0:AUTH: $addr found in $list/opt.");
				return 1;
			    }elsif($m eq "# $ml"){
				$e{'poster'} = $addr;
				&Log("$0:AUTH: $addr found in $list(off)/opt.");
				return 2;
			    }
			}
		    }else{ 
			$e{'poster'} = $addr;
			&Log("$0:AUTH: $addr($members{$addr}) found in $list.");
			return 1;
		    }
		}elsif($exist == 2){
		    $e{'poster'} = $addr;
		    &Log("$0:AUTH: $addr found in $list(off).");
		    return 2;
		}elsif ($exist == 3){
		    if ($members{$lcaddr}[1] ne ""){
			for ($i = 1; $members{$lcaddr}[$i] ne ""; $i++){
			    $m = $members{$lcaddr}[$i];
			    if ($m eq $ml){
				$e{'poster'} = $addr;
				&Log("$0:AUTH: $addr found in $list/opt/old.");
				return 1;
			    }elsif($m eq "# $ml"){
				$e{'poster'} = $addr;
				&Log("$0:AUTH: $addr found in $list(off)/opt/old.");
				return 2;
			    }
			}
		    }else{ 
			$e{'poster'} = $addr;
			&Log("$0:AUTH: $addr($members{$lcaddr}) found in $list.");
			return 1;
		    }
		}elsif($exist == 4){
		    $e{'poster'} = $addr;
		    &Log("$0:AUTH: $addr found in $list(off/old).");
		    return 2;

		}
	    }
	}
    }
    $e{'poster'} = $ret[0];
    &Log("$0:Fail: $addr not found in $list.") if ($debug > 1);
    return 0;
}

sub Conv2MBs{
    local($ent,*ret) = @_;
    local($atom,$addr);
    $ent =~ s/\n//g;
    while (length($ent) > 0){
	$addr = "";
	&Log("$0:ENT:$ent") if ($debug > 5);
	if ($ent =~ /^\s*([^\<\>]*)\s*\<(\S+)\>[\s\,]*(.*)/){
	    $atom++; $addr = $2; $addr = &FormatAddr($addr); $addr =~ s/[\s]//g;
	    push(@ret,$addr);
	    $ent = $3;
	}elsif($ent =~ /^\s*([^\(\)\s]+)\s*\(([^\)]*)\)[\s\,]*(.*)/){
	    $atom++; $addr = $1; $addr = &FormatAddr($addr); $addr =~ s/[\s]//g;
	    push(@ret,$addr);
	    $ent = $3;
	}elsif($ent =~ /^\s*([^\(\)\<\>\s]+)[\s\,]*(.*)/){
	    $atom++; $addr = $1; $addr = &FormatAddr($addr); $addr =~ s/[\s]//g;
	    push(@ret,$addr);
	    $ent = $2;
	}else{
	    $ent = "";
	}
    }
    if ($atom == 0) { return 0; };
    return 1;
}

sub MIMEdecode{
    $0 = "MIMEdecode";
    local(*e) = @_;
    local($dc) = 0;
    eval("require(\"mimer.pl\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
	exit $EX_CONFIG;
    }
    foreach $hh (keys(%e)){
	if ( $hh =~ /^H\:/ ){
	    if ( $e{$hh} =~ /\=\?ISO\-2022\-JP\?B\?/i ){
		$e{$hh} = &mimedecode($e{$hh});
		$dc++;
	    }
	}
    }
    &Log("$0:MIME: Decode $dc headers.");
    return 1;
}

sub MIMEencode{
    $0 = "MIMEencode";
    local(*e) = @_;
    local($ec)= 0;
    eval("require(\"mimew.pl\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
	exit $EX_SOFTWARE;
    }
    foreach $hh (keys(%e)){
	if ( $hh =~ /^H\:/ ){
	    if ( $e{$hh} =~ /\033../ ){
		$e{$hh} = &mimeencode($e{$hh});
		$ec++;
	    }
	}
    }
    &Log("$0:MIME: Encode $ec headers.");
    return 1;
}


sub IncrSeqNum{
    $0 = "IncrSeqNum";
    local(*e) = @_;
    if ( -f $Seqfile ){
	open(SEQ,$Seqfile);
	$seq = <SEQ>;
	chop($seq);
	close(SEQ);
	$seq = int($seq);
    }else{
	$seq = 0;
    }
    ### JUST INCREMENT :)
    $seq++;
    $e{'seq'} = $seq;
    open(SEQ,">$Seqfile");
    print SEQ "$seq\n";
    close(SEQ);
    return 1;
}

sub DecrSeqNum{
    $0 = "DecrSeqNum";
    local(*e) = @_;
    if ( -f $Seqfile ){
	open(SEQ,$Seqfile);
	$seq = <SEQ>;
	chop($seq);
	close(SEQ);
	$seq = int($seq);
    }else{
	$seq = 0;
    }
    ### JUST INCREMENT :)
    $seq--;
    $e{'seq'} = $seq;
    open(SEQ,">$Seqfile");
    print SEQ "$seq\n";
    close(SEQ);
    return 1;
}

sub SaveAsFile{
    $0 = "SaveAsFile";
    local(*e) = @_;
    local($wait) = 0;
    if (!defined($e{'seq'})){ 
	&Log("$0:FILE: Skip saving because \$e{'seq'} is not defined.");
	return 1;
    }
    if (-f "$Savedir/$e{'seq'}"){
	rename("$Savedir/$e{'seq'}","$Savedir/$e{'seq'}.bak");
	&Log("$0:FILE: $Savedir/$e{'seq'} is already exist. saved as .bak");
    }
    while(!open(SF,">$Savedir/$e{'seq'}")){
	sleep 3;
	$wait++;
	if ($wait > 100){
	    &Log("$0:Fail: Timeout to open $Savedir/$e{'seq'}");
	    print STDERR "Fail: Timeout to save file\n";
	    exit $EX_IOERR;
	}
    }
    foreach $hh (sort(keys(%e))){
	if ($hh =~ /^H\:(\S+)$/){
	    $h = $e{"OH:$1"} || $1;
	    print SF "$h\: $e{$hh}";
	}
    }
    print SF "\n";
    print SF "$e{'Body'}";
    close(SF);
    &Log("$0:FILE: Article saved as $Savedir/$e{'seq'}");
    return 1;
}

sub DeletePart{
    $0 = "DeletePart";
    local(*e,$b,$i) = @_;
    local($tb);
    local($mcb) = sprintf("mp:%02d",$b);
    local($mcs) = sprintf("mp:%02d.%02d",$b,$i);
    if (($e{'mp:00'}) && ($b <= $e{'mp:00'}) && ($i <= $e{$mcb})){
	$tb = substr($e{'Body'},0,index($e{'Body'},$e{$mcs}));
	$tb .= substr($e{'Body'},index($e{'Body'},$e{$mcs})+$e{"size:$mcs"});
	$e{'Body'} = $tb;
	&Log("$0:MIME: delete $mcs part succeed");
	return 1;
    }else{
	&Log("$0:MIME: can not find $mcs part");
	return 0;
    }
}

sub Distribute{
    $0 = "Distribute";
    local(*e) = @_;
    local(%r);
    local(%ddp);
    local($rem,$hhh,$hhhh);
    local(@rcpts,@dmls,@mls);
    local(@rrcpts) = ();
    local($rc)  = 0;
    local($tc)  = 0;
    local($toc) = 0;
    local($total) = 0;
    local($soc) = 0;
    local($cont,$tb);
    local($body) = $e{'Body'};
    local($ml,$domain) = split(/\@/,$ML_ADDR);
    $SESSIONMAXRCPTS = $SESSIONMAXRCPTS || 100; # default
    $domain =~ s/\s//g;
    $ml =~ s/\s//g;
    if (!defined($e{'mp:00'})){
	if (defined($MAXBODYSIZE) && ($e{'size:body'} > $MAXBODYSIZE)){
	    ### body size check
	    foreach $kk (keys(%e)){
		if ($kk =~ /^H:Content/){
		    delete $e{$kk};
		}
            }
	    $e{'H:Content-Type'} = "text/plain; charset=iso-2022-jp\n";
	    $e{'H:MIME-Version'} = "1.0\n";
	    $tb = "\n\nThe size ($e{'size:body'}) exceeds the limit.\n";
	    $tb .= "If you need original, ";
	    $tb .= "Please send the line below to $ML_CONTROL,\n\n";
	    $tb .= "\tget $ml $e{'seq'}\n\n--$ML_ADMIN\n";
	    $body = $tb;
	}
    }elsif($e{'mp:00'} > 0){
	### part size check
	for ($b = 0; $b <= $e{'mp:00'} ; $b++){
	    local($mcb) = sprintf("mp:%02d",$b);
	    for ($i = 1; $i <= $e{$mcb} ; $i++){
		local($mcs) = sprintf("mp:%02d.%02d",$b,$i);
		if (defined($MAXPARTSIZE) && ($e{"size:$mcs"} > $MAXPARTSIZE)){
		    $tb = substr($e{'Body'},0,index($e{'Body'},$e{"$mcs"}));
		    $tb .= "--";
		    $tb .= $e{"$mcs:boundary"};
		    $tb .= "\n";
		    $tb .= "Content-Type: Text/Plain; charset=iso-2022-jp\n\n\n";
		    $tb .= "The size (";
		    $tb .= $e{"size:$mcs"};
		    $tb .= "bytes) exceeds the limit.\n";
		    $tb .= "If you need original, ";
		    $tb .= "Please send the line below to $ML_CONTROL,\n\n";
		    $tb .= "\tget $ml $e{'seq'}\n\n--$ML_ADMIN\n";
		    $tb .= substr($e{'Body'},index($e{'Body'},$e{$mcs})+$e{"size:$mcs"});
		    $body = $tb;
		}
	    }
	}
    }
    ### for Group-Server extension
    if ($OPTIMIZED_DELIVERLIST){
	&Conv2MBs($e{'H:To'},*rcpts);
	&Conv2MBs($e{'H:Cc'},*rcpts);
	foreach $ad (@rcpts){
	    ($lp,$dp) = split(/\@/,$ad);
	    $dp =~ s/\,//g;
            $lp =~ s/\,//g;
	    &Log("$0:OPTD:ML'$lp'\@'$dp'");
	    if ((-d "$ML_SETTING_ROOT/$lp")||($lp eq $ml)||($MLALIAS{$lp} eq $ml)){
                if (defined($MLALIAS{$lp})){
                    push(@mls,$MLALIAS{$lp});
                    &Log("$0:OPTD:ML=$MLALIAS{$lp} ($lp)");
                }else{
                    push(@mls,$lp);
                    &Log("$0:OPTD:ML=$lp");
                }
	    }elsif(($dp eq $domain) && ($OPTIMIZED_SKIP)){
		$ddp{"$lp\@$dp"} = 1;
		&Log("$0:OPTD:SKIP=$lp\@$dp");
	    }
	}
    }
    ###
    if ($SENDMAIL){
	if (!defined(@DELIVERLIST)){
	    $cont = "You should define \@DELIVERLIST in config.ph.\n";
	    $cont .= "You can't deliver with \@DELI*B*ERLIST(;.;).\n";
	    $cont .= "(This typo fixed in ppml-0.16)\n";
	    &NoticeMail($ML_ADMIN,'@DELIVERLIST not defined',$cont);
	    return 0;
	}
	foreach $list (@DELIVERLIST){
	    delete $INC{$list};
	    eval("require(\"$list\")");
	    if ($@) {
		&Log("$0:ERROR:$@");
		print STDERR $@;
		exit $EX_CONFIG;
	    }
	    foreach $addr (keys(%members)){
		if ($addr !~ /^\#/){
		    ### for Group-Server extension
		    if ($OPTIMIZED_DELIVERLIST){
			undef @dmls;
			for ($i = 1; $members{$addr}[$i] ne ""; $i++){
			    $op = $members{$addr}[$i];
			    foreach $ad (@mls){
				if ($ad eq $op){
				    push(@dmls,$ad);
				    &Log("$0:ML:$ad") if ($debug > 1);
				}
			    }
			}
			if ($dmls[0] eq $ml){
			    unless ($ddp{$addr}){
				&Log("$0:RCPT: $addr: $dmls[0] $dmls[1] $dmls[2]");
				$LIST .= " $addr";
				$toc++;
			    }
			}
		    }else{
			&Log("$0:RCPT: $addr") if ($debug > 1);
			$LIST .= " $addr";
			$toc++;
		    }
		}
		if (length($LIST) > 1800){
		    open(SC,"|$SENDMAIL -f $ML_ADMIN -- $LIST");
		    foreach $hh (@H_ATOM){
			$hhh = "H\:$hh";
			$hhhh = $e{"OH:$hh"} || $hh;
			if ( $e{$hhh} ){
			    &Log("$0:HEAD: $hhhh\: $e{$hhh}") if ($debug > 8);
			    print SC "$hhhh\: $e{$hhh}";
			    $r{$hhh} = $e{$hhh};
			    delete $e{$hhh};
			}
		    }
		    foreach $hh (sort(keys(%e))){
			if ($hh =~ /^H\:(\S+)$/){
			    $h = $e{"OH:$1"} || $1;
			    &Log("$0:HEAD: $h\: $e{$hh}") if ($debug > 8);
			    print SC "$h\: $e{$hh}";
			}
		    }
		    print SC "\n";
		    print SC "$body\n";
		    close(SC);
	            $LIST = "";
		    foreach (keys(%r)){
			$e{$_} = $r{$_};
			delete $r{$_};
		    }
		    $soc++;
		}
	    }
	    if (length($LIST) > 0){
		open(SC,"|$SENDMAIL -f $ML_ADMIN -- $LIST");
		foreach $hh (@H_ATOM){
		    $hhh = "H\:$hh";
		    $hhhh = $e{"OH:$hh"} || $hh;
		    if ( $e{$hhh} ){
			&Log("$0:HEAD: $hhhh\: $e{$hhh}") if ($debug > 8);
			print SC "$hhhh\: $e{$hhh}";
			$r{$hhh} = $e{$hhh};
			delete $e{$hhh};
		    }
		}
		foreach $hh (sort(keys(%e))){
		    if ($hh =~ /^H\:(\S+)$/){
			$h = $e{"OH:$1"} || $1;
			&Log("$0:HEAD: $h\: $e{$hh}") if ($debug > 8);
			print SC "$h\: $e{$hh}";
		    }
		}
		print SC "\n";
		print SC "$body\n";
		close(SC);
		$LIST = "";
		foreach (keys(%r)){
		    $e{$_} = $r{$_};
		    delete $r{$_};
		}
		$soc++;
	    }
	}
	&Log("$0:PIPE: Put ${toc} rcpts in ${soc} transaction");
	return 1;
    }
    $body =~ s/\n\./\n\.\./g;
    foreach $list (@DELIVERLIST){
	delete $INC{$list};
	eval("require(\"$list\")");
	if ($@) {
	    &Log("$0:ERROR:$@");
	    print STDERR $@;
	    exit $EX_CONFIG;
	}
	foreach $addr (keys(%members)){
	    if ($addr !~ /^\#/){
		if ($OPTIMIZED_DELIVERLIST){
		    undef @dmls;
		    for ($i = 1; $members{$addr}[$i] ne ""; $i++){
			$op = $members{$addr}[$i];
			foreach $ad (@mls){
			    &Log("$0:OPTD:TRY($op/$ad)") if ($debug > 5);
			    if ($ad eq $op){
				push(@dmls,$ad);
				&Log("$0:OPTD:ML=OK($op/$ad)") if ($debug > 5);
			    }
			}
		    }
		    if ($dmls[0] eq $ml){
			unless ($ddp{$addr}){
			    push(@rrcpts,$addr);
			    $tc++;
			    &Log("$0:RCPT: $addr: $dmls[0] $dmls[1] $dmls[2] $dmls[3]");
			}
		    }
		}else{
		    push(@rrcpts,$addr);
		    $tc++;
		}
		if ($tc >= $SESSIONMAXRCPTS) {
		    $toc += $tc;
		    $total += &InetDeliver;
		    $soc++;
		    $tc = 0;
		    undef(@rrcpts);
		}
	    }
	}
    }
    if ($tc > 0){
	$toc += $tc;
	$total += &InetDeliver;
	$soc++;
	$tc = 0;
	undef(@rrcpts);
    }
    &Log("$0:SMTP: Sent ${total}/${toc} rcpts in ${soc} transactions.");
    return 1;
}


sub InetDeliver{
    $0 = "InetDeliver";
    local($rc) = 0;
    while (!$rem){
	$rem = IO::Socket::INET->new(Proto =>"tcp",PeerAddr => $SMTP_HOST,);
    }
    $rem->autoflush(1);
    $resp = <$rem>; &Log("$0:SMTP:$conn: $resp") if ($debug > 1);
    print $rem "HELO $PPML_HOST\r\n";
    $resp = <$rem>; &Log("$0:SMTP:$conn: $resp") if ($debug > 1);
    print $rem "Mail From:$ML_ADMIN\r\n";
    $resp = <$rem>; &Log("$0:SMTP:$conn: $resp") if ($debug > 1);
    foreach $rcpt (@rrcpts){
	print $rem "Rcpt To:<$rcpt>\r\n";
	$resp = <$rem>; 
	if ($resp =~ /^250/){
	    &Log("$0:SMTP:$conn: $resp") if ($debug > 1);
	    $rc++;
	}else{
	    &Log("$0:SMTP:$conn: $resp");
	}
    }
    if ($rc > 0){
        &Log("$0:SMTP:$conn: $BRACKET $e{'seq'} : $rc rcpts OK.");
    }else{
        &Log("$0:Fail:$conn: $BRACKET $e{'seq'} : No ($rc) rcpts.");
        print $rem "QUIT\r\n";
        undef $rem;
        return 1;
    }
    print $rem "Data\r\n";
    $resp = <$rem>; &Log("$0:SMTP:$conn: $resp") if ($debug > 1);
    foreach $hh (@H_ATOM){
	$hhh = "H\:$hh";
	$hhhh = $e{"OH:$hh"} || $hh;
	if ( $e{$hhh} ){
	    $rnl = $e{$hhh};
	    $rnl =~ s/([^\r])\n/$1\r\n/g;
	    print $rem "$hhhh\: $rnl";
	    $r{$hhh} = $e{$hhh};
	    delete $e{$hhh};
	}
    }
    foreach $hh (sort(keys(%e))){
	if ($hh =~ /^H\:(\S+)$/){
	    $h = $e{"OH:$1"} || $1;
	    print $rem "$h\: $e{$hh}";
	}
    }
    print $rem "\r\n";
    $body =~ s/([^\r])\n/$1\r\n/g;
    print $rem "$body\r\n.\r\n";
    $resp = <$rem>; &Log("$0:SMTP:$conn: $resp");
    print $rem "QUIT\r\n";
    $resp = <$rem>; &Log("$0:SMTP:$conn: $resp") if ($debug > 1);
    close($rem);
    undef $rem;
    foreach (keys(%r)){
	$e{$_} = $r{$_};
	delete $r{$_};
    }
    return $rc;
}

sub NoticeMail{
    $0 = "NoticeMail";
    local($rcpt,$sub,$cont,*r) = @_;
    local($rem);
    local($rc) = 0;
    ####
    $r{'H:Subject'} = $sub;
    $r{'H:From'} = "PPML-Report <$ML_ADMIN>";
    $r{'H:To'} = $rcpt;
    ####
    if ($SENDMAIL){
	open(SC,"|$SENDMAIL -f $ML_ADMIN -- $rcpt");
	foreach $hh (keys(%r)){
	    if ($hh =~ /^H\:(\S+)$/){
		print SC "$1\: $r{$hh}\r\n";
	    }
	}
	print SC "\n";
	print SC "$cont\n";
	close(SC);
	&Log("$0:PIPE: Put notice to $SENDMAIL");
	return 1;
    }
    #### SMTP
    $cont =~ s/\n\./\n\.\./g;
    while (!$rem){
	$rem = IO::Socket::INET->new(Proto =>"tcp",PeerAddr => $SMTP_HOST,);
	sleep 3;
    }
    $rem->autoflush(1);
    $resp = <$rem>; &Log("$0:SMTP: $resp") if ($debug > 1);
    print $rem "HELO $PPML_HOST\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp") if ($debug > 1);
    print $rem "Mail From:$ML_ADMIN\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp") if ($debug > 1);
    print $rem "Rcpt To:<$rcpt>\r\n";
    $resp = <$rem>; 
    if ($resp =~ /^250/){
	$rc++;
	&Log("$0:SMTP: $resp") if ($debug > 1);
    }else{
	&Log("$0:SMTP: $resp");
    }
    if ($rc > 0){
	&Log("$0:SMTP: $BRACKET $e{'seq'} : $rc rcpts OK.");
    }else{
	&Log("$0:Fail: $BRACKET $e{'seq'} : $rc rcpts.");
	exit $EX_OK;
    }
    print $rem "Data\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp") if ($debug > 1);
    foreach $hh (keys(%r)){
	if ($hh =~ /^H\:(\S+)$/){
	    print $rem "$1\: $r{$hh}\r\n";
	}
    }
    print $rem "\r\n";
    $rnl = $cont;
    $rnl =~ s/([^\r])\n/$1\r\n/g;
    print $rem "$rnl\r\n.\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp");
    print $rem "QUIT\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp");
    return  1;
}

sub FormatAddr{
    $0 = "FormatAddr";
    my($addr) = @_;
    local($lp,$dp) = split(/\@/,$addr);
    $dp =~ tr/A-Z/a-z/;
    $addr = $lp.'@'.$dp;
    return $addr;
}

sub LcAddr{
    $0 = "LcAddr";
    my($addr) = @_;
    $addr =~ tr/A-Z/a-z/;
    return $addr;
}

sub PPLockWait{
    $0 = "PPLockWait";
    local($wait) = 0;
    unless(-d "$Savedir"){
	mkdir("$Savedir",0750);
	&Log("$0:FILE: \$Savedir($Savedir) is created.");
    }
    if ($USE_LOCKFILE == 1){
	while (-f "$Savedir/.lockfile"){
	    sleep 1;
	    $wait++;
	    if ($wait > 15){
		&Log("$0:Fail: Timeout to wait unlock ($Savedir/.lockfile)");
		print STDERR "$0:Fail: Timeout to wait unlock ($Savedir/.lockfile)\n";
		exit $EX_TEMPFAIL;
	    }
	}
	open (LOCK, "> $Savedir/$locktmp");
	print LOCK "$$\n";
	close (LOCK);
	while (!link("$Savedir/$locktmp", "$Savedir/.lockfile")){
	    sleep 1;
	    $wait++;
	    if ($wait > 32){
		&Log("$0:Fail: Timeout to link ($Savedir/.lockfile)");
		print STDERR "$0:Fail: Timeout to link ($Savedir/.lockfile)\n";
                unlink("$Savedir/$locktmp");
		exit $EX_TEMPFAIL;
	    }
	}
	&Log("$0:LOCK: lock Lockfile($Savedir/.lockfile) -- session start");
    }else{
	while (!open(LOCK,"+< config.ph")){
	    sleep 1;
	    $wait++;
	    if ($wait > 15){
		&Log("$0:Fail: Timeout to open (config.ph)");
		print STDERR "$0:Fail: Timeout to open (config.ph)\n";
		exit $EX_TEMPFAIL;
	    }
	}
	while (!flock(LOCK,$LOCK_EX|$LOCK_NB)){
	    sleep 1;
	    $wait++;
	    if ($wait > 32){
		&Log("$0:Fail: Timeout to lock (config.ph)");
		print STDERR "$0:Fail: Timeout to lock (config.ph)\n";
		exit $EX_TEMPFAIL;
	    }
	}
	&Log("$0:LOCK: lock config.ph -- session start");
    }
    return 1;
}

sub Unlock{
    $0 = "Unlock";
    &Log("$0:LOCK: unlock config.ph -- number generated");
    if ($USE_LOCKFILE == 1){
	unlink("$Savedir/.lockfile");
	unlink("$Savedir/$locktmp");
    }else{
	unless (flock(LOCK,$LOCK_UN)){
	    &Log("$0:Fail: Can't unlock (config.ph)") if $debug;
	}
	close(LOCK);
    }
    return 1;
}

sub Done{
    $0 = "Done";
    &Log("$0:DONE: OK -- session completed.");
    exit $EX_OK;
}


sub DebugPrint{
    local(*e) = @_;
    foreach $key (sort(keys(%e))){
	print "-------- $key:\n$e{$key}\n";
    }
    return 1;
}
    
sub Log{
    local($s) = @_;
    ($sec,$min,$hour,$mday,$mon,$year,$wday) = (localtime(time))[0..6];
    $now = sprintf("%04d.%02d.%02d %02d:%02d:%02d",
		   1900 + $year, $mon + 1, $mday, $hour, $min ,$sec);
    $s =~ s/\r//;
    $s =~ s/\n//;
    open(LOG,">>$Logfile");
    print LOG "$now $s\n";
    close(LOG);
    return 1;
}

